'use strict';

function wasmWrapper (worker, functions) {
  // We need to make sure that we connect to 'our' worker
  // TODO: this is gross, think of another way
  const key = '@@symbol:__wasm_api__';
  const secret = Math.random().toString(36).slice(2);
  const api = Object.create(null);
  const ready = handleMessage((resolve, { data }, cb) => {
    if (data[key] !== secret) return

    detachCallback(cb);
    resolve();
  });

  let pingPongMsgCounter = 0;

  // worker.postMessage(functions)

  for (const { name, args: argTypes, ret } of functions) {
    const returnsArray = /[0-9]/.test(ret);

    api[name] = returnsArray
      ? fnThatReturnsArray(name, argTypes.length)
      : (...args) => call(name, args);
  }

  worker.postMessage({
    [key]: secret,
    functions
  });

  function handleMessage (callback) {
    return new Promise(resolve => {
      worker.addEventListener('message', function cb (e) {
        callback(resolve, e, cb);
      });
    })
  }

  function fnThatReturnsArray (name, declaredLength) {
    return function (...args) {
      if (!args.length || args.length !== declaredLength + 1) {
        throw new Error(
          `[wasm-client]: failed to call the '${name}' function. ` +
          'A function that returns an array should accept the array\'s length ' +
          'as the last parameter.'
        )
      }

      const arrayLength = args.pop();

      if (typeof arrayLength !== 'number') {
        throw new TypeError(
          '[wasm-worker]: arrayLength must be a number'
        )
      }

      return call(name, args, arrayLength)
    }
  }

  function detachCallback (cb) {
    worker.removeEventListener('message', cb);
  }

  function call (name, args, arrayLength) {
    const id = `${secret}-${pingPongMsgCounter}`;

    const promise = handleMessage((resolve, { data }, cb) => {
      // Is this check necessary?
      if (data.id !== id || data.name !== name) return

      detachCallback(cb);
      pingPongMsgCounter++;
      resolve(data.ret);
    });

    const transferables = [];
    args && args.length && args.forEach(arg => {
      !!arg.buffer && transferables.push(arg.buffer);
    });

    worker.postMessage({ id, name, args, arrayLength }, transferables);

    return promise
  }

  return ready.then(() => api)
}

module.exports = wasmWrapper;
